/* -*- mode: Java; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Copyright (c) 2011-2011 Laird Nelson.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * The original copy of this license is available at
 * http://www.opensource.org/license/mit-license.html.
 */
package liquibase.logging.ext;

import java.util.Collections;
import java.util.EnumMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import java.util.logging.Level;
import java.util.logging.Logger;

import liquibase.logging.LogLevel;

import liquibase.logging.core.AbstractLogger;
import liquibase.logging.core.DefaultLogger; // for javadoc only

import liquibase.servicelocator.PrioritizedService; // for javadoc only

/**
 * An {@link AbstractLogger} that adapts the Java logging mechanism to
 * the <a href="http://liquibase.org">Liquibase</a> {@linkplain
 * liquibase.logging.Logger logging contracts}.
 *
 * @author <a href="mailto:ljnelson@gmail.com">Laird Nelson</a>
 *
 * @version 1.0
 *
 * @since 1.0
 */
public class JavaUtilLoggingLogger extends AbstractLogger {

  /**
   * An immutable {@link Map} of <a
   * href="http://liquibase.org">Liquibase</a> {@link LogLevel}
   * instances to {@link java.util.logging.Level} instances.  This
   * field is never {@code null} and is threadsafe.
   */
  public static final Map<LogLevel, Level> levels;

  /**
   * Static initializer; initializes the {@link #levels} field.
   */
  static {
    final Map<LogLevel, Level> levelMap = new EnumMap<LogLevel, Level>(LogLevel.class);
    levelMap.put(LogLevel.OFF, Level.OFF);
    levelMap.put(LogLevel.DEBUG, Level.FINER);
    levelMap.put(LogLevel.INFO, Level.INFO);
    levelMap.put(LogLevel.WARNING, Level.WARNING);
    levelMap.put(LogLevel.SEVERE, Level.SEVERE);
    levels = Collections.unmodifiableMap(levelMap);
  }

  /**
   * The underlying {@link Logger} that this {@link
   * JavaUtilLoggingLogger} delegates to.  This field may be {@code
   * null} at any point.
   */
  private transient Logger logger;

  /**
   * Creates a new {@link JavaUtilLoggingLogger}.  Following <a
   * href="http://liquibase.org/">Liquibase</a> convention, callers
   * should immediately call {@link #setName(String)} after
   * construction.
   */
  public JavaUtilLoggingLogger() {
    super();
  }
  
  /**
   * Returns {@code 3} when invoked.
   *
   * @see PrioritizedService#getPriority()
   */
  @Override
  public int getPriority() {
    return 3; // arbitrary number greater than 1
  }
  
  /**
   * Sets this {@link JavaUtilLoggingLogger}'s log level to the
   * supplied {@link LogLevel} and sets its underlying {@link Logger}
   * implementation's {@linkplain Logger#getLevel() level} as well.
   *
   * @param level the new {@link LogLevel}; may be {@code null} in
   * which case {@link Level#OFF} will be used instead
   */
  @Override
  public void setLogLevel(final LogLevel level) {
    super.setLogLevel(level);
    if (this.logger != null && level != null) {
      final Level l = levels.get(level);
      if (l == null) {
        this.logger.setLevel(Level.OFF);
      } else {
        this.logger.setLevel(l);
      }
    }
  }

  /**
   * Emulates the behavior of <a
   * href="http://liquibase.org">Liquibase</a>'s {@link
   * DefaultLogger#setLogLevel(String, String)} method by simply
   * ignoring the second argument.  This method calls {@link
   * AbstractLogger#setLogLevel(String)}.
   *
   * @param logLevel the new log level; may be {@code null}
   *
   * @param fileName ignored; the {@linkplain AbstractLogger
   * superclass} does not document what this parameter is to be used
   * for or what its permissible values are
   */
  @Override
  public void setLogLevel(final String logLevel, final String fileName) {
    this.setLogLevel(logLevel); // emulates DefaultLogger behavior
  }
  
  /**
   * Returns the {@link LogLevel} that was previously {@linkplain
   * #setLogLevel(LogLevel) set} on this {@link
   * JavaUtilLoggingLogger}.
   *
   * <p>This implementation actually reverse engineers the {@link
   * LogLevel} by inspecting the {@link Logger#getLevel()} of the
   * underlying Java {@link Logger}, and using the {@link #levels}
   * {@link Map} to locate the proper {@link LogLevel}.  This allows
   * subclasses or future versions of this class to safely expose the
   * underlying {@link Logger} if deemed necessary.</p>
   *
   * <p>This method may return {@code null}.</p>
   *
   * @return the {@link LogLevel} used by this {@link
   * JavaUtilLoggingLogger} instance, or {@code null}
   */
  @Override
  public LogLevel getLogLevel() {
    if (this.logger == null) {
      return super.getLogLevel();
    }
    final Level level = this.logger.getLevel();
    if (level == null) {
      return LogLevel.OFF;
    }
    final Set<Entry<LogLevel, Level>> entrySet = levels.entrySet();
    assert entrySet != null;
    assert !entrySet.isEmpty();
    for (final Entry<LogLevel, Level> entry : entrySet) {
      assert entry != null;
      if (level.equals(entry.getValue())) {
        return entry.getKey();
      }
    }
    throw new IllegalStateException();
  }
  
  /**
   * Sets the name of this {@link JavaUtilLoggingLogger}.
   *
   * <p>This method will initialize the underlying {@link Logger} with
   * a call to the {@link Logger#getLogger(String)} method, passing
   * the supplied {@code name} parameter.</p>
   *
   * @param name the name; may be {@code null}
   */
  @Override
  public void setName(final String name) {
    this.logger = Logger.getLogger(name);
  }

  /**
   * Returns the {@link StackTraceElement} representing the closest
   * method invocation that is <em>not</em> an invocation of a method
   * in this class, its superclasses or {@link java.lang.Thread}.
   *
   * <p>This method is used by others in this class to find the class
   * name and method name to pass to the underlying Java {@link
   * Logger}s when actually performing logging.</p>
   *
   * <p>This method may return {@code null}.</p>
   *
   * @return the nearest valid {@link StackTraceElement}, or {@code
   * null}
   */
  private final StackTraceElement getCaller() {
    StackTraceElement returnValue = null;
    // TODO: doPrivileged
    final StackTraceElement[] st = Thread.currentThread().getStackTrace();
    if (st != null && st.length > 0) {
      for (final StackTraceElement element : st) {
        if (element != null && element.getMethodName() != null) {
          final String className = element.getClassName();
          if (className != null && !className.equals("java.lang.Thread") && !className.equals(this.getClass().getName())) {
            returnValue = element;
            break;
          }
        }
      }
    }
    return returnValue;
  }
  
  /**
   * Calls the {@link #info(String, Throwable)} method, passing {@code
   * null} as the second argument.
   *
   * @param message the message to log; may be {@code null}
   */
  @Override
  public void info(final String message) {
    this.info(message, null);
  }
  
  /**
   * Logs the supplied message {@link String} and (optional) {@link
   * Throwable} to the underlying {@link Logger} at the {@link
   * Level#INFO INFO} level.
   *
   * @param message the message to log; may be {@code null}
   *
   * @param throwable the {@link Throwable} to log; may be {@code
   * null}
   */
  @Override
  public void info(final String message, final Throwable throwable) {
    if (this.logger != null && this.logger.isLoggable(Level.INFO)) {
      final StackTraceElement ste = this.getCaller();
      if (throwable == null) {        
        this.logger.logp(Level.INFO, ste.getClassName(), ste.getMethodName(), message);
      } else {
        this.logger.logp(Level.INFO, ste.getClassName(), ste.getMethodName(), message, throwable);
      }
    }
  }
  
  /**
   * Calls the {@link #warning(String, Throwable)} method, passing {@code
   * null} as the second argument.
   *
   * @param message the message to log; may be {@code null}
   */
  @Override
  public void warning(final String message) {
    this.warning(message, null);
  }
  
  /**
   * Logs the supplied message {@link String} and (optional) {@link
   * Throwable} to the underlying {@link Logger} at the {@link
   * Level#WARNING WARNING} level.
   *
   * @param message the message to log; may be {@code null}
   *
   * @param throwable the {@link Throwable} to log; may be {@code
   * null}
   */
  @Override
  public void warning(final String message, final Throwable throwable) {
    if (this.logger != null && this.logger.isLoggable(Level.WARNING)) {
      final StackTraceElement ste = this.getCaller();
      if (throwable == null) {
        this.logger.logp(Level.WARNING, ste.getClassName(), ste.getMethodName(), message);
      } else {
        this.logger.logp(Level.WARNING, ste.getClassName(), ste.getMethodName(), message, throwable);
      }
    }
  }

  /**
   * Calls the {@link #severe(String, Throwable)} method, passing {@code
   * null} as the second argument.
   *
   * @param message the message to log; may be {@code null}
   */
  @Override
  public void severe(final String message) {
    this.severe(message, null);
  }

  /**
   * Logs the supplied message {@link String} and (optional) {@link
   * Throwable} to the underlying {@link Logger} at the {@link
   * Level#SEVERE SEVERE} level.
   *
   * @param message the message to log; may be {@code null}
   *
   * @param throwable the {@link Throwable} to log; may be {@code
   * null}
   */
  @Override
  public void severe(final String message, final Throwable throwable) {
    if (this.logger != null && this.logger.isLoggable(Level.SEVERE)) {
      final StackTraceElement ste = this.getCaller();
      if (throwable == null) {
        this.logger.logp(Level.SEVERE, ste.getClassName(), ste.getMethodName(), message);
      } else {
        this.logger.logp(Level.SEVERE, ste.getClassName(), ste.getMethodName(), message, throwable);
      }
    }
  }

  /**
   * Calls the {@link #debug(String, Throwable)} method, passing {@code
   * null} as the second argument.
   *
   * @param message the message to log; may be {@code null}
   */
  @Override
  public void debug(final String message) {
    this.debug(message, null);
  }

  /**
   * Logs the supplied message {@link String} and (optional) {@link
   * Throwable} to the underlying {@link Logger} at the {@link
   * Level#FINER FINER} level.
   *
   * @param message the message to log; may be {@code null}
   *
   * @param throwable the {@link Throwable} to log; may be {@code
   * null}
   */
  @Override
  public void debug(final String message, final Throwable throwable) {
    if (this.logger != null && this.logger.isLoggable(Level.FINER)) {
      final StackTraceElement ste = this.getCaller();
      if (throwable == null) {
        this.logger.logp(Level.FINER, ste.getClassName(), ste.getMethodName(), message);
      } else {
        this.logger.logp(Level.FINER, ste.getClassName(), ste.getMethodName(), message, throwable);
      }
    }
  }
  

}